package org.ioc.commons.impl.gwt.client.flowcontrol.operationmanager;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ioc.commons.flowcontrol.common.BindRegistration;
import org.ioc.commons.flowcontrol.operationmanager.IsOperation;
import org.ioc.commons.flowcontrol.operationmanager.OperationHandler;
import org.ioc.commons.flowcontrol.operationmanager.OperationManager;

import com.google.web.bindery.event.shared.HandlerRegistration;

public class GwtOperationManagerImpl<O extends Enum<? extends IsOperation>> implements OperationManager<O> {

	private static class GwtOperationManagerBindRegistration implements BindRegistration {

		private HandlerRegistration operationBeginningHandlerRegistration;
		private HandlerRegistration operationFinishedHandlerRegistration;

		public GwtOperationManagerBindRegistration(HandlerRegistration operationBeginningHandlerRegistration,
				HandlerRegistration operationFinishedHandlerRegistration) {
			this.operationBeginningHandlerRegistration = operationBeginningHandlerRegistration;
			this.operationFinishedHandlerRegistration = operationFinishedHandlerRegistration;
		}

		@Override
		public void unbind() {
			this.operationBeginningHandlerRegistration.removeHandler();
			this.operationFinishedHandlerRegistration.removeHandler();
		}

	}

	private com.google.web.bindery.event.shared.EventBus eventBus;
	private List<O> onGoingOps = new ArrayList<O>();
	private Map<OperationHandler<O>, BindRegistration> onceHandlers = new HashMap<OperationHandler<O>, BindRegistration>();

	public GwtOperationManagerImpl(com.google.web.bindery.event.shared.EventBus eventBus) {
		this.eventBus = eventBus;
	}

	@Override
	public BindRegistration bindOperation(final O operation, final OperationHandler<O> handler) {

		HandlerRegistration operationBeginningHandlerRegistration = this.eventBus.addHandler(
				OperationBeginningEvent.TYPE, new OperationBeginningEvent.Handler() {

					@Override
					public void onOperationBeginning(
							OperationBeginningEvent<? extends Enum<? extends IsOperation>> event) {
						if (event.getOperation().equals(operation)) {
							handler.onOperationBeginning(event.getSource(), operation);
						}
					}
				});

		HandlerRegistration operationFinishedHandlerRegistration = this.eventBus.addHandler(
				OperationFinishedEvent.TYPE, new OperationFinishedEvent.Handler() {

					@Override
					public void onOperationFinished(OperationFinishedEvent<? extends Enum<? extends IsOperation>> event) {
						if (event.getOperation().equals(operation)) {
							handler.onOperationFinished(event.getSource(), operation, event.isSuccess());

							if (!onceHandlers.isEmpty()) {
								BindRegistration registration = onceHandlers.remove(handler);
								if (registration != null) {
									registration.unbind();
								}
							}
						}
					}
				});

		GwtOperationManagerBindRegistration registration = new GwtOperationManagerBindRegistration(
				operationBeginningHandlerRegistration, operationFinishedHandlerRegistration);

		return registration;
	}

	@Override
	public BindRegistration bindOperationOnce(final O operation, final OperationHandler<O> handler) {

		BindRegistration registration = bindOperation(operation, handler);

		this.onceHandlers.put(handler, registration);

		return registration;
	}

	@Override
	public BindRegistration bindOperationsRunning(final OperationHandler<O> handler) {
		HandlerRegistration operationBeginningHandlerRegistration = this.eventBus.addHandler(
				OperationBeginningEvent.TYPE, new OperationBeginningEvent.Handler() {

					@SuppressWarnings("unchecked")
					@Override
					public void onOperationBeginning(
							OperationBeginningEvent<? extends Enum<? extends IsOperation>> event) {
						if (onGoingOps.size() == 1) {
							handler.onOperationBeginning(event.getSource(), (O) event.getOperation());
						}
					}
				});

		HandlerRegistration operationFinishedHandlerRegistration = this.eventBus.addHandler(
				OperationFinishedEvent.TYPE, new OperationFinishedEvent.Handler() {

					@SuppressWarnings("unchecked")
					@Override
					public void onOperationFinished(OperationFinishedEvent<? extends Enum<? extends IsOperation>> event) {
						if (onGoingOps.isEmpty()) {
							handler.onOperationFinished(event.getSource(), (O) event.getOperation(), event.isSuccess());
						}
					}
				});

		GwtOperationManagerBindRegistration registration = new GwtOperationManagerBindRegistration(
				operationBeginningHandlerRegistration, operationFinishedHandlerRegistration);

		return registration;
	}

	@Override
	public void operationBeginning(O operation) {
		this.onGoingOps.add(operation);
		eventBus.fireEvent(new OperationBeginningEvent<Enum<? extends IsOperation>>(operation));
	}

	@Override
	public boolean operationFinished(O operation, boolean success) {
		boolean removed = this.onGoingOps.remove(operation);
		eventBus.fireEvent(new OperationFinishedEvent<Enum<? extends IsOperation>>(operation, success));
		return removed;
	}

	@Override
	public int getPendingOps() {
		return onGoingOps.size();
	}

	@Override
	public boolean isBusy() {
		return (!this.onGoingOps.isEmpty());
	}

	@Override
	public boolean isOpPending(O operation) {
		return onGoingOps.contains(operation);
	}
}
