package org.ioc.commons.impl.gwt.integration.client.dataaccess.requestfactory.rebind.internal;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.validation.ConstraintViolation;

import org.ioc.commons.integration.dataaccess.dao.IsDaoFactory;
import org.ioc.commons.integration.dataaccess.dao.Tx;
import org.ioc.commons.utils.Format;
import org.ioc.commons.utils.FormatterLogger;

import com.google.web.bindery.requestfactory.shared.BaseProxy;
import com.google.web.bindery.requestfactory.shared.Receiver;
import com.google.web.bindery.requestfactory.shared.RequestContext;
import com.google.web.bindery.requestfactory.shared.RequestFactory;
import com.google.web.bindery.requestfactory.shared.ServerFailure;

public abstract class RFTxRequestProvider<D extends IsDaoFactory, F extends RequestFactory> implements Tx<D>, RequestProvider<F> {

	private static FormatterLogger logger = FormatterLogger.getLogger(RFTxRequestProvider.class);

	protected RequestContext mainContext;
	protected final F dataFactory;

	protected final String transactionName;
	protected final Map<String, RequestQueue> queueRequests;
	protected final Map<BaseProxy, Tx<D>> transactionMap;

	protected boolean canBeFired;

	/*
	 * FIXME: Cuando esté resuelto el Issue 5525, este parámetro ya no tendrá
	 * sentido y se podrá eliminar. Otra opción es dejarlo sin usar "por si".
	 */
	private int pendingOps;

	private boolean pendingFire;

	public RFTxRequestProvider(F dataFactory, String transactionName, Map<String, RequestQueue> queueRequests,
			Map<BaseProxy, Tx<D>> transactionMap) {
		this.dataFactory = dataFactory;
		this.transactionName = transactionName;
		this.queueRequests = queueRequests;
		this.transactionMap = transactionMap;

		this.canBeFired = true;

		logger.info("Starting transaction '{0}'", this);
	}

	@Override
	public abstract D daoFactory();

	public final String getTransactionName() {
		return transactionName;
	}

	protected final <T extends RequestContext> T appendNewRequest(T request) {
		if (this.mainContext == null) {
			this.mainContext = request;
			logger.info("Estableciendo mainContext para la transacción '{0}': {1}", this, this.mainContext);
		} else {
			this.mainContext.append(request);
			logger.info("... añadiendo context en la transacción '{0}': {1} (MainContext={2})", this, request,
					this.mainContext);
		}
		return request;
	}

	public final RequestContext getMainContext() {
		return mainContext;
	}

	@Override
	public final void fireTransaction() {

		if (this.pendingOps > 0) {
			logger.info("Dejando transacción '{0}' en espera hasta que finalicen {1} operaciones pendientes...", this,
					this.pendingOps);
			this.pendingFire = true;
			return;
		}

		if (this.mainContext != null) {
			RFDaoReceiver<Void> receiver = new RFDaoReceiver<Void>() {

				@Override
				public void success(Void response) {
					super.success(response);
					removeAllProxiesBound();
				}

				@Override
				public void failure(ServerFailure grupitServerFailure) {
					super.failure(grupitServerFailure);
					removeAllProxiesBound();
				}

				@Override
				public void constraintViolation(Set<ConstraintViolation<?>> violations) {
					super.constraintViolation(violations);

					/**
					 * Si se produce error de violación, RequestFactory permite
					 * volver a ser lanzado
					 */
					canBeFired = true;
				}

			};
			logger.info("Lanzando transacción '{0}'...", this);
			canBeFired = false;
			this.mainContext.fire(receiver);

			/**
			 * Aquí hay un problema por diseño de GWT:
			 * <ul>
			 * <li>onSucess deja el RC no valido (no se puede hacer otro fire)</li>
			 * <li>
			 * onConstraintViolation lo deja válido (se puede hacer otro fire)</li>
			 * <li>El problema onFailure que es ambiguo. Hay 2 tipos de errores,
			 * unos que llaman genéricos que lo deja válido y otros que lo dejan
			 * invalido. El problema es que no hay forma de determinar cuando se
			 * ha invocado cual. Por tanto vamos marcar sólo como no valido
			 * cuando termine el onSuccess y esperemos no tener problemas.</li>
			 * </ul>
			 * Si no se produce un contraintValidation, esta transacción ya no
			 * es válida. El propio sistema de GWT se encargará al realizar el
			 * fire() de que pete.
			 */

			logger.info("... transacción '{0}' lanzada.", this);
		}
	}

	@Override
	public final void fireOrQueueTransaction(String queueName) {
		if (this.mainContext != null) {
			RequestQueue rq = queueRequests.get(queueName);

			if (rq == null) {
				rq = new RequestQueue();
				queueRequests.put(queueName, rq);
			}

			Receiver<Void> receiver = new Receiver<Void>() {

				@Override
				public void onConstraintViolation(Set<ConstraintViolation<?>> violations) {
					super.onConstraintViolation(violations);
					canBeFired = true;
				}

				@Override
				public void onSuccess(Void response) {
					removeAllProxiesBound();
				}

				@Override
				public void onFailure(ServerFailure error) {
					super.onFailure(error);
					removeAllProxiesBound();
				}

			};
			canBeFired = false;
			logger.info("Encolando transacción '{0}' en '{1}'...", this, queueName);
			rq.fireOrQueue(this.mainContext);
			logger.info("... transacción '{0}' encolada en '{1}'.", this, queueName);
		}
	}

	@Override
	public final boolean canBeFired() {
		return canBeFired;
	}

	@Override
	public final boolean getAutoFire() {
		return false;
	}

	@Override
	public final F getDataFactory() {
		return this.dataFactory;
	}

	public final void pendingOpStart() {
		pendingOps++;
	}

	public final void pendingOpEnd() {
		--pendingOps;

		if (pendingOps <= 0 && this.pendingFire) {
			this.pendingFire = false;
			logger.info("Operaciones pendientes terminadas. Lanzamos transacción...");
			this.fireTransaction();
		}
	}

	@Override
	public String toString() {
		return Format.substitute("[{0}]{1}", this.transactionName, super.toString());
	}

	final public void bindProxy(BaseProxy proxy) {
		this.transactionMap.put(proxy, this);
	}

	private void removeAllProxiesBound() {
		/**
		 * Borramos todos los proxies asociados
		 */
		Set<Entry<BaseProxy, Tx<D>>> es = this.transactionMap.entrySet();
		for (Iterator<Entry<BaseProxy, Tx<D>>> it = es.iterator(); it.hasNext();) {
			Entry<BaseProxy, Tx<D>> et = it.next();
			if (et.getValue().equals(this)) {
				it.remove();
			}
		}
		// for (BaseProxy p : this.transactionMap.keySet()) {
		// if (this.transactionMap.get(p).equals(this)) {
		// this.transactionMap.remove(p);
		// }
		// }
	}

}
