package org.ioc.commons.integration.dataaccess.dao;

import org.ioc.commons.flowcontrol.operationmanager.IsOperation;
import org.ioc.commons.flowcontrol.operationmanager.OperationManagerTrigger;

/**
 * Manager for an operation performed through a DAO.
 * 
 * @author Jesús Lunar Pérez
 * 
 * @param <T>
 *            Received data type
 */
public interface DaoOperationManager<T> {
	/**
	 * Specifies the objet which will receive the data
	 * 
	 * @param receiver
	 *            The receiver
	 * @return An operation manager for continuing managing the operation.
	 */
	DaoOperationManager<T> to(DaoReceiver<T> receiver);

	/**
	 * Specifies the minimum required properties (accesible attributes of the T
	 * class through getters) to be filled in the T object when it is received.
	 * 
	 * @param propertyRefs
	 *            A set of properfies from T
	 * @return An operation manager for continuing managing the operation.
	 */
	DaoOperationManager<T> with(String... propertyRefs);

	/**
	 * Put this operation in a named queue.
	 * 
	 * The operations put on a named queue will be executed consecutively as a
	 * FIFO queue. A queue is created internally when is used the first time.
	 * 
	 * @param queueName
	 *            Name of the queue
	 * @return An operation manager for continuing managing the operation.
	 */
	DaoOperationManager<T> queue(String queueName);

	/**
	 * Handle the beginning and the end of the operation through an
	 * {@link OperationManagerTrigger}
	 * 
	 * @param operation
	 *            Operation to be managed during the DAO operation.
	 * @param trigger
	 *            Trigger for manage the beginning and the end of the operation.
	 * @return An operation manager for continuing managing the operation.
	 */
	<O extends Enum<? extends IsOperation>> DaoOperationManager<T> performing(O operation,
			OperationManagerTrigger<O> trigger);
}
